module moon_master_top(clk, rst_n, PORT_A, PORT_B, PORT_C, PORT_D
					, hsync, vsync, RGB_out, line_int, frame_int);

input clk;
input rst_n;

input [7:0] PORT_A;
input [7:0] PORT_B;
input [7:0] PORT_C;
input [7:0] PORT_D;

reg [7:0] PORT_A_reg;
reg [7:0] PORT_B_reg;
reg [7:0] PORT_C_reg;
reg [7:0] PORT_D_reg;

output hsync, vsync, line_int, frame_int;

output reg [7:0] RGB_out;

// tile RAM controls
wire tRAM_wren;
wire tRAM_wren_DT;

wire [7:0] tRAM_data;

wire [9:0] tRAM_addr;
wire [9:0] tRAM_addr_DT;
reg [9:0] tRAM_addr_VGA;


//sprite RAM controls
wire spRAM_wren;
wire spRAM_wren_DT;

wire [7:0] spRAM_data;

wire [7:0] spRAM_addr;
wire [7:0] spRAM_addr_DT;
wire [7:0] spRAM_addr_read;


//screen RAM controls

wire scRAM_wren;
wire scRAM_wren_DT;

wire [3:0] scRAM_data;

wire [8:0] scRAM_addr;
wire [8:0] scRAM_addr_DT;
reg [8:0] scRAM_addr_VGA;


// end controls

wire [7:0] row;
wire [7:0] col;

wire [7:0] col_off;

wire [7:0] tRAM_out;
wire [7:0] spRAM_out;
wire [3:0] scRAM_out;

wire RAM_to_screen;  //true if we're currently blasting RAM to the screen, false otherwise

wire sprite_en;

wire draw_sprite;

wire [1:0] mode;

reg [2:0] xoff;
wire scroll;
wire reset_xoff;

//register scroll
reg last_scroll;
wire posedge_scroll;

assign posedge_scroll = scroll & ~last_scroll;

//slow clock has 1/3 freq of clk
reg slowclk;
reg [1:0] slowclkcnt;

assign mode = PORT_D[7:6];
assign scroll = PORT_C[7];
assign reset_xoff = PORT_C[6];

assign sprite_en = (mode == 2'b00) ? 1'b1 : 1'b0;

assign tRAM_wren = (RAM_to_screen == 1'b1) ? 1'b0: tRAM_wren_DT;
assign scRAM_wren = (RAM_to_screen == 1'b1) ? 1'b0: scRAM_wren_DT;
assign spRAM_wren = (RAM_to_screen == 1'b1) ? 1'b0: spRAM_wren_DT;

assign tRAM_addr = (RAM_to_screen == 1'b1) ? tRAM_addr_VGA : tRAM_addr_DT;

assign scRAM_addr = (RAM_to_screen == 1'b1) ? scRAM_addr_VGA : scRAM_addr_DT;

assign spRAM_addr = (RAM_to_screen == 1'b1) ? spRAM_addr_read : spRAM_addr_DT;

assign col_off = col + xoff;

data_transfer idata_transfer(clk, rst_n, PORT_A_reg, PORT_B_reg, PORT_C_reg, PORT_D_reg
					, tRAM_wren_DT, tRAM_data, tRAM_addr_DT
					, spRAM_wren_DT, spRAM_data, spRAM_addr_DT
					, scRAM_wren_DT, scRAM_data, scRAM_addr_DT);			

tRAM itRAM(clk, tRAM_wren, tRAM_addr, tRAM_data, tRAM_out);

spRAM ispRAM(clk, spRAM_wren, spRAM_addr, spRAM_data, spRAM_out);

scRAM iscRAM(clk, scRAM_wren, scRAM_addr, scRAM_data, scRAM_out);

sprite_draw isprite_draw(slowclk, rst_n, sprite_en, PORT_A[7:0]
						, PORT_B[7:0], PORT_C[1:0], PORT_C[3:2], row, col, spRAM_addr_read, draw_sprite);

vga_out ivga_out(slowclk, rst_n, hsync, vsync, frame_int, line_int, RAM_to_screen, row, col);


//register input ports
always @(posedge clk)
if (rst_n == 1'b0)
begin
	PORT_A_reg <= 8'b0;
	PORT_B_reg <= 8'b0;
	PORT_C_reg <= 8'b0;
	PORT_D_reg <= 8'b0;
end
else
begin
	PORT_A_reg <= PORT_A;
	PORT_B_reg <= PORT_B;
	PORT_C_reg <= PORT_C;
	PORT_D_reg <= PORT_D;
end

//create slow clock
always @(posedge clk)
if (rst_n == 1'b0)
begin
	slowclkcnt <= 2'd0;
	slowclk <= 1'b0;
end
else
begin
	if (slowclkcnt == 2'b10)
	begin
		slowclkcnt <= 2'b00;
	end
	else
	begin
		slowclkcnt <= slowclkcnt + 2'd1;
	end
	
	if (slowclkcnt == 2'b10)
	begin
		slowclk <= 1'b0;
	end
	else if (slowclkcnt == 2'b01)
	begin
		slowclk <= 1'b1;
	end
end

always @(posedge clk)
if (rst_n == 1'b0)
begin
	xoff <= 3'd0;
	last_scroll <= 0;
end
else
begin
	last_scroll <= scroll;

	if (posedge_scroll == 1'b1)
		xoff <= xoff + 3'd1;
	else if (reset_xoff)
		xoff <= 3'd0;
end


always @(draw_sprite, spRAM_out, tRAM_out, RAM_to_screen)
begin
	if (RAM_to_screen == 1'b0)
		RGB_out <= 0;
	else if (draw_sprite == 1'b0)
		RGB_out <= tRAM_out;
	else
	begin
		if (spRAM_out == 8'hFE)
			RGB_out <= tRAM_out;
		else
			RGB_out <= spRAM_out;
	end
end


always @(posedge slowclk)
if (rst_n == 1'b0)
begin
	scRAM_addr_VGA <= 0;
	tRAM_addr_VGA <= 0;
end
else
begin
	scRAM_addr_VGA <= col_off[7:3] + 9'd26*row[7:3]; 
	tRAM_addr_VGA <= {scRAM_out[3:0], row[2:0], col_off[2:0]};
end

endmodule